# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import pytest
log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def remove_redshift_conn():
    if 'REDSHIFT_CONN' in os.environ:
        del os.environ['REDSHIFT_CONN']
    if 'WRITE_SCHEMA' in os.environ:
        del os.environ['WRITE_SCHEMA']


@pytest.fixture(scope="function")
def we_module_env():
    os.environ['REDSHIFT_CONN'] = 'postgresql://user:password@host.io:1234/db'
    os.environ['WRITE_SCHEMA'] = 'testing'


@pytest.fixture(scope="function")
def dagbag():
    log.debug("dagbag()")
    os.environ['AIRFLOW__CORE__SQL_ALCHEMY_CONN'] = 'sqlite://'
    os.environ['MARQUEZ_NAMESPACE'] = 'test-marquez'

    from airflow import settings
    import airflow.utils.db as db_utils
    db_utils.resetdb(settings.RBAC)
    from airflow.models import DagBag
    dagbag = DagBag(include_examples=False)
    return dagbag



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

log = logging.getLogger(__name__)


class MockIdMapping:

    mapping = None

    def __init__(self):
        log.debug("MockIdMapping.init()")
        self.mapping = {}

    def set(self, key, value):
        self.mapping[key] = value

    def pop(self, key, _session):
        return self.mapping.pop(key)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from airflow.operators.postgres_operator import PostgresOperator

from marquez_airflow.utils import get_connection_uri
from marquez_airflow.extractors.sql.experimental.parser import SqlParser
from marquez_airflow.extractors import (
    BaseExtractor,
    StepMetadata,
    Source,
    Dataset
)

log = logging.getLogger(__name__)


class PostgresExtractor(BaseExtractor):
    operator_class = PostgresOperator

    def __init__(self, operator):
        super().__init__(operator)

    def extract(self) -> [StepMetadata]:
        # (1) Parse sql statement to obtain input / output tables.
        sql_meta = SqlParser.parse(self.operator.sql)
        log.info("postgres sql parse successful.")

        # (2) Default all inputs / outputs to current connection.
        # NOTE: We'll want to look into adding support for the `database`
        # property that is used to override the one defined in the connection.
        conn_id = self.operator.postgres_conn_id
        source = Source(
            type='POSTGRESQL',
            name=conn_id,
            connection_url=get_connection_uri(conn_id))

        # (3) Map input / output tables to dataset objects with source set
        # as the current connection. NOTE: We may want to move _to_dataset()
        # to class BaseExtractor or a utils class.
        inputs = [
            Dataset.from_table(source, table) for table in sql_meta.in_tables
        ]
        outputs = [
            Dataset.from_table(source, table) for table in sql_meta.out_tables
        ]

        return [StepMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=outputs,
            context={
                'sql': self.operator.sql
            }
        )]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import time
from uuid import uuid4

from airflow import LoggingMixin
from airflow.contrib.operators.bigquery_operator import BigQueryOperator
from pendulum import Pendulum

import airflow.models
from airflow.operators.postgres_operator import PostgresOperator

from marquez_airflow.extractors.bigquery_extractor import BigQueryExtractor
from marquez_airflow.utils import (
    JobIdMapping,
    get_location,
    add_airflow_info_to
)
from marquez_airflow.extractors import (Dataset, Source, StepMetadata)
from marquez_airflow.extractors.postgres_extractor import PostgresExtractor

from marquez_client import Clients
from marquez_client.models import JobType

_NOMINAL_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

_DAG_DEFAULT_NAMESPACE = 'default'
_DAG_DEFAULT_OWNER = 'anonymous'


class DAG(airflow.models.DAG, LoggingMixin):
    DEFAULT_NAMESPACE = 'default'
    _job_id_mapping = None
    _marquez_client = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._marquez_dataset_cache = {}
        self._marquez_source_cache = {}
        self.marquez_namespace = os.getenv('MARQUEZ_NAMESPACE',
                                           _DAG_DEFAULT_NAMESPACE)
        self._job_id_mapping = JobIdMapping()
        # TODO: Manually define operator->extractor mappings for now,
        # but we'll want to encapsulate this logic in an 'Extractors' class
        # with more convenient methods (ex: 'Extractors.extractor_for_task()')
        self._extractors = {
            PostgresOperator: PostgresExtractor,
            BigQueryOperator: BigQueryExtractor
            # Append new extractors here
        }
        self.log.debug(
            f"DAG successfully created with extractors: {self._extractors}"
        )

    def create_dagrun(self, *args, **kwargs):
        # run Airflow's create_dagrun() first
        dagrun = super(DAG, self).create_dagrun(*args, **kwargs)

        create_dag_start_ms = self._now_ms()
        execution_date = kwargs.get('execution_date')
        run_args = {
            'external_trigger': kwargs.get('external_trigger', False)
        }

        # Marquez metadata collection
        try:
            # TODO: Use 'anonymous' owner for now, but we may want to use
            # the 'owner' attribute defined via default_args for a DAG
            self.log.debug(
                f"Creating namespace '{self.marquez_namespace}' with "
                f"owner '{_DAG_DEFAULT_OWNER}'...")
            self.get_or_create_marquez_client()\
                .create_namespace(self.marquez_namespace, _DAG_DEFAULT_OWNER)

            # Register each task in the DAG
            for task_id, task in self.task_dict.items():
                t = self._now_ms()
                try:
                    self.report_task(
                        dagrun.run_id,
                        execution_date,
                        task,
                        self._extractors.get(task.__class__),
                        run_args=run_args)
                except Exception as e:
                    self.log.error(
                        f'Failed to record task: {e} '
                        f'airflow_dag_id={self.dag_id} '
                        f'task_id={task_id} '
                        f'marquez_namespace={self.marquez_namespace} '
                        f'duration_ms={(self._now_ms() - t)}')

            self.log.info(
                f'Successfully recorded metadata: '
                f'airflow_dag_id={self.dag_id} '
                f'marquez_namespace={self.marquez_namespace} '
                f'duration_ms={(self._now_ms() - create_dag_start_ms)}')

        except Exception as e:
            self.log.error(
                f'Failed to record metadata: {e} '
                f'airflow_dag_id={self.dag_id} '
                f'marquez_namespace={self.marquez_namespace} '
                f'duration_ms={(self._now_ms() - create_dag_start_ms)}')

        return dagrun

    def handle_callback(self, *args, **kwargs):
        try:
            dagrun = args[0]
            task_instances = dagrun.get_task_instances()
            for ti in task_instances:
                extractor = self._extractors.get(ti.operator.__class__)
                if extractor:
                    self.log.debug(
                        f"Using extractor '{extractor}' to apply partial "
                        f"metadata updates for operator '{ti.operator}' "
                        f"in DAG '{self.dag_id}'."
                    )

                    steps_meta = add_airflow_info_to(
                        ti.operator,
                        extractor(ti.operator).extract_on_complete(ti)
                    )
                    self.log.debug(f'steps_meta: {steps_meta}')

                    # FIXME: replace with update api
                    for step in steps_meta:
                        self.get_marquez_client().create_job(
                            job_name=step.name,
                            job_type=JobType.BATCH,
                            location=(step.location or self._get_location(ti)),
                            input_dataset=step.inputs,
                            output_dataset=step.outputs,
                            context=step.context,
                            description=self.description,
                            namespace_name=self.marquez_namespace)

                try:
                    self.report_jobrun_change(
                        ti, dagrun.run_id, **kwargs)
                except Exception as e:
                    self.log.error(
                        f'Failed to record task run state change: {e} '
                        f'dag_id={self.dag_id}')

        except Exception as e:
            self.log.error(
                f'Failed to record dagrun state change: {e} '
                f'dag_id={self.dag_id}')

        return super().handle_callback(*args, **kwargs)

    def report_task(self,
                    dag_run_id,
                    execution_date,
                    task,
                    extractor,
                    marquez_job_run_id=None,
                    run_args=None):
        self.log.info("report_task()")
        report_job_start_ms = self._now_ms()
        marquez_client = self.get_or_create_marquez_client()
        if execution_date:
            start_time = self._to_iso_8601(execution_date)
            end_time = self.compute_endtime(execution_date)
        else:
            start_time = None
            end_time = None

        if end_time:
            end_time = self._to_iso_8601(end_time)

        task_location = None
        try:
            if hasattr(task, 'file_path') and task.file_path:
                task_location = get_location(task.file_path)
            else:
                task_location = get_location(task.dag.fileloc)
        except Exception as e:
            self.log.warning(
                f'Unable to fetch the location: {e}')

        steps_metadata = []
        if extractor:
            try:
                self.log.info(
                    f'Using extractor {extractor.__name__} '
                    f'task_type={task.__class__.__name__} '
                    f'airflow_dag_id={self.dag_id} '
                    f'task_id={task.task_id} '
                    f'airflow_run_id={dag_run_id} '
                    f'marquez_namespace={self.marquez_namespace}')
                steps_metadata = add_airflow_info_to(
                    task,
                    extractor(task).extract()
                )
            except Exception as e:
                self.log.error(
                    f'Failed to extract metadata {e} '
                    f'airflow_dag_id={self.dag_id} '
                    f'task_id={task.task_id} '
                    f'airflow_run_id={dag_run_id} '
                    f'marquez_namespace={self.marquez_namespace}')
        else:
            self.log.warning(
                f'Unable to find an extractor. '
                f'task_type={task.__class__.__name__} '
                f'airflow_dag_id={self.dag_id} '
                f'task_id={task.task_id} '
                f'airflow_run_id={dag_run_id} '
                f'marquez_namespace={self.marquez_namespace}')

        task_name = f'{self.dag_id}.{task.task_id}'

        # If no extractor found or failed to extract metadata,
        # report the task metadata
        if not steps_metadata:
            steps_metadata = add_airflow_info_to(
                task,
                [StepMetadata(name=task_name)]
            )

        # store all the JobRuns associated with a task
        marquez_jobrun_ids = []

        for step in steps_metadata:
            self.log.info(f'step: {step}')
            inputs = []
            try:
                self.register_datasets(step.inputs)
                inputs = list(
                    map(lambda input: {
                        'namespace': self.marquez_namespace,
                        'name': input.name
                    }, step.inputs))
                self.log.info(f'inputs: {inputs}')
            except Exception as e:
                self.log.error(
                    f'Failed to register inputs: {e} '
                    f'inputs={str(step.inputs)} '
                    f'airflow_dag_id={self.dag_id} '
                    f'task_id={task.task_id} '
                    f'step={step.name} '
                    f'airflow_run_id={dag_run_id} '
                    f'marquez_namespace={self.marquez_namespace}')

            outputs = []
            try:
                self.register_datasets(step.outputs, marquez_job_run_id)
                outputs = list(
                    map(lambda output: {
                        'namespace': self.marquez_namespace,
                        'name': output.name
                    }, step.outputs))
                self.log.info(f'outputs: {outputs}')
            except Exception as e:
                self.log.error(
                    f'Failed to register outputs: {e} '
                    f'outputs={str(step.outputs)} '
                    f'airflow_dag_id={self.dag_id} '
                    f'task_id={task.task_id} '
                    f'step={step.name} '
                    f'airflow_run_id={dag_run_id} '
                    f'marquez_namespace={self.marquez_namespace}')

            marquez_client.create_job(job_name=step.name,
                                      job_type=JobType.BATCH,  # job type
                                      location=(step.location or
                                                task_location),
                                      input_dataset=inputs,
                                      output_dataset=outputs,
                                      context=step.context,
                                      description=self.description,
                                      namespace_name=self.marquez_namespace)
            self.log.info(
                f'Successfully recorded job: {step.name} '
                f'airflow_dag_id={self.dag_id} '
                f'marquez_namespace={self.marquez_namespace}')

            # NOTE: When we have a run ID generated by Marquez, skip creating
            # a new run as we only want to update the dataset with the existing
            # run ID
            if marquez_job_run_id:
                return

            # TODO: Look into generating a uuid based on the DAG run_id
            external_run_id = str(uuid4())

            marquez_client.create_job_run(
                namespace_name=self.marquez_namespace,
                job_name=step.name,
                run_id=external_run_id,
                run_args=run_args,
                nominal_start_time=start_time,
                nominal_end_time=end_time)

            if external_run_id:
                marquez_jobrun_ids.append(external_run_id)
                marquez_client.mark_job_run_as_started(run_id=external_run_id)
            else:
                self.log.error(
                    f'Failed to get run id: {step.name} '
                    f'airflow_dag_id={self.dag_id} '
                    f'airflow_run_id={dag_run_id} '
                    f'marquez_namespace={self.marquez_namespace}')
            self.log.info(
                f'Successfully recorded job run: {step.name} '
                f'airflow_dag_id={self.dag_id} '
                f'airflow_dag_execution_time={start_time} '
                f'marquez_run_id={external_run_id} '
                f'marquez_namespace={self.marquez_namespace} '
                f'duration_ms={(self._now_ms() - report_job_start_ms)}')

        # Store the mapping for all the steps associated with a task
        try:
            self._job_id_mapping.set(
                JobIdMapping.make_key(task_name, dag_run_id),
                json.dumps(marquez_jobrun_ids))

        except Exception as e:
            self.log.error(
                f'Failed to set id mapping : {e} '
                f'airflow_dag_id={self.dag_id} '
                f'task_id={task.task_id} '
                f'airflow_run_id={dag_run_id} '
                f'marquez_run_id={marquez_jobrun_ids} '
                f'marquez_namespace={self.marquez_namespace}')

    def compute_endtime(self, execution_date):
        return self.following_schedule(execution_date)

    def report_jobrun_change(self, ti, run_id, **kwargs):
        job_name = f'{ti.dag_id}.{ti.task_id}'
        session = kwargs.get('session')
        marquez_job_run_ids = self._job_id_mapping.pop(
            JobIdMapping.make_key(job_name, run_id), session)

        if marquez_job_run_ids:
            self.log.info(
                f'Found job runs: '
                f'airflow_dag_id={self.dag_id} '
                f'airflow_job_id={job_name} '
                f'airflow_run_id={run_id} '
                f'marquez_run_ids={marquez_job_run_ids} '
                f'marquez_namespace={self.marquez_namespace}')

            ids = json.loads(marquez_job_run_ids)
            if kwargs.get('success'):
                for marquez_job_run_id in ids:
                    for task_id, task in self.task_dict.items():
                        if task_id == ti.task_id:
                            self.log.info(f'task_id: {task_id}')
                            self.report_task(
                                run_id,
                                None,
                                task,
                                self._extractors.get(task.__class__),
                                marquez_job_run_id=marquez_job_run_id)
                    self.get_or_create_marquez_client(). \
                        mark_job_run_as_completed(run_id=marquez_job_run_id)
            else:
                for marquez_job_run_id in ids:
                    self.get_or_create_marquez_client().mark_job_run_as_failed(
                        run_id=marquez_job_run_id)

        state = 'COMPLETED' if kwargs.get('success') else 'FAILED'
        self.log.info(
            f'Marked job run(s) as {state}. '
            f'airflow_dag_id={self.dag_id} '
            f'airflow_job_id={job_name} '
            f'airflow_run_id={run_id} '
            f'marquez_run_id={marquez_job_run_ids} '
            f'marquez_namespace={self.marquez_namespace}')

    def get_or_create_marquez_client(self):
        if not self._marquez_client:
            self._marquez_client = Clients.new_write_only_client()
        return self._marquez_client

    @staticmethod
    def _now_ms():
        return int(round(time.time() * 1000))

    def register_datasets(self, datasets, marquez_job_run_id=None):
        client = self.get_or_create_marquez_client()
        for dataset in datasets:
            if isinstance(dataset, Dataset):
                _key = f'{self.marquez_namespace}.{dataset.name}'
                if _key not in self._marquez_dataset_cache:
                    self.register_source(dataset.source)
                    client.create_dataset(
                        dataset_name=dataset.name,
                        dataset_type=dataset.type,
                        physical_name=dataset.name,
                        source_name=dataset.source.name,
                        namespace_name=self.marquez_namespace,
                        run_id=marquez_job_run_id)
                    # NOTE:
                    self._marquez_dataset_cache[_key] = True

    def register_source(self, source):
        if isinstance(source, Source):
            _key = source.name
            if _key not in self._marquez_source_cache:
                client = self.get_or_create_marquez_client()
                client.create_source(
                    source.name,
                    source.type,
                    source.connection_url)
                # NOTE:
                self._marquez_source_cache[_key] = True

    @staticmethod
    def _to_iso_8601(dt):
        if isinstance(dt, Pendulum):
            return dt.format(_NOMINAL_TIME_FORMAT)
        else:
            return dt.strftime(_NOMINAL_TIME_FORMAT)

    @staticmethod
    def _get_location(task):
        try:
            if hasattr(task, 'file_path') and task.file_path:
                return get_location(task.file_path)
            else:
                return get_location(task.dag.fileloc)
        except Exception as e:
            DAG.log.warning(f'Unable to fetch the location. {e}')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from abc import ABC, abstractmethod

from airflow.models import BaseOperator

from marquez_client.models import DatasetType

log = logging.getLogger(__name__)


class Source:
    name = None
    connection_url = None
    type = None

    def __init__(self, name, type, connection_url):
        self.name = name
        self.type = type
        self.connection_url = connection_url

    def __eq__(self, other):
        return self.name == other.name and \
               self.type == other.type and \
               self.connection_url == other.connection_url

    def __repr__(self):
        return f"Source({self.name!r}, {self.type!r}, {self.connection_url!r})"


class Dataset:
    source: Source = None
    name = None
    description = None
    type = None

    def __init__(self, source, name, type, description=None):
        self.source = source
        self.name = name
        self.type = type
        self.description = description

    @staticmethod
    def from_table(source, table):
        return Dataset(
            type=DatasetType.DB_TABLE,
            name=table,
            source=source
        )

    def __eq__(self, other):
        return self.source == other.source and \
               self.name == other.name and \
               self.type == other.type and \
               self.description == other.description

    def __repr__(self):
        return f"""
            Dataset(\
              {self.source!r}, \
              {self.name!r}, \
              {self.type!r}, \
              {self.description!r})
        """


class StepMetadata:
    # TODO: Define a common way across extractors to build the
    # job name for an operator
    name = None
    location = None
    inputs = []
    outputs = []
    context = {}

    def __init__(self, name, location=None, inputs=None, outputs=None,
                 context=None):
        self.name = name
        self.location = location
        if inputs:
            self.inputs = inputs
        if outputs:
            self.outputs = outputs
        if context:
            self.context = context

    def __repr__(self):
        return "name: {}\t inputs: {} \t outputs: {}".format(
            self.name,
            ','.join([str(i) for i in self.inputs]),
            ','.join([str(o) for o in self.outputs]))


class BaseExtractor(ABC):
    operator: BaseOperator = None
    operator_class = None

    def __init__(self, operator):
        log.debug("BaseExtractor.init")
        self.operator = operator

    @classmethod
    def get_operator_class(cls):
        return cls.operator_class

    def validate(self):
        # TODO: maybe we should also enforce the module
        assert (self.operator_class is not None and
                self.operator.__class__ == self.operator_class)

    # TODO: Only return a single StepMetadata object.
    @abstractmethod
    def extract(self) -> [StepMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> [StepMetadata]:
        # TODO: This method allows for the partial updating of task
        # metadata on completion. Marquez currently doesn't support
        # partial updates within the context of a DAG run, but this feature
        # will soon be supported:
        # https://github.com/MarquezProject/marquez/issues/816
        #
        # Also, we'll want to revisit the metadata extraction flow,
        # but for now, return an empty set as the default behavior
        # as not all extractors need to handle partial metadata updates.
        return []



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import mock

from airflow.models import (TaskInstance, DagRun)
from airflow.operators.dummy_operator import DummyOperator
from airflow.utils.db import provide_session
from airflow.utils.dates import days_ago
from airflow.utils import timezone
from airflow.utils.state import State

from marquez_client.models import JobType

from marquez_airflow import DAG
from marquez_airflow.utils import get_location

NO_INPUTS = []
NO_OUTPUTS = []

DEFAULT_DATE = timezone.datetime(2016, 1, 1)

DAG_ID = 'test_dag'
DAG_RUN_ID = 'test_run_id_for_task_completed_and_failed'
DAG_RUN_ARGS = {'external_trigger': False}
DAG_NAMESPACE = 'default'
DAG_OWNER = 'anonymous'
DAG_DESCRIPTION = \
    'A simple DAG to test the marquez.DAG metadata extraction flow.'

DAG_DEFAULT_ARGS = {
    'owner': DAG_OWNER,
    'depends_on_past': False,
    'start_date': days_ago(1),
    'email_on_failure': False,
    'email_on_retry': False,
    'email': ['owner@test.com']
}

DAG = DAG(
    DAG_ID,
    schedule_interval='@daily',
    default_args=DAG_DEFAULT_ARGS,
    description=DAG_DESCRIPTION
)

TASK_ID_COMPLETED = 'test_task_completed'
TASK_ID_FAILED = 'test_task_failed'


@pytest.fixture
@provide_session
def clear_db_airflow_dags(session=None):
    session.query(DagRun).delete()
    session.query(TaskInstance).delete()


@mock.patch('marquez_airflow.DAG.get_or_create_marquez_client')
@provide_session
def test_marquez_dag(mock_get_or_create_marquez_client,
                     clear_db_airflow_dags, session=None):
    # (1) Mock the marquez client method calls
    mock_marquez_client = mock.Mock()
    mock_get_or_create_marquez_client.return_value = mock_marquez_client

    # (2) Add task that will be marked as completed
    task_will_complete = DummyOperator(
        task_id=TASK_ID_COMPLETED,
        dag=DAG
    )
    completed_task_location = get_location(task_will_complete.dag.fileloc)

    # (3) Add task that will be marked as failed
    task_will_fail = DummyOperator(
        task_id=TASK_ID_FAILED,
        dag=DAG
    )
    failed_task_location = get_location(task_will_complete.dag.fileloc)

    # (4) Create DAG run and mark as running
    dagrun = DAG.create_dagrun(
        run_id=DAG_RUN_ID,
        execution_date=DEFAULT_DATE,
        state=State.RUNNING)

    # Assert namespace meta call
    mock_marquez_client.create_namespace.assert_called_once_with(DAG_NAMESPACE,
                                                                 DAG_OWNER)

    # Assert source and dataset meta calls
    mock_marquez_client.create_datasource.assert_not_called()
    mock_marquez_client.create_dataset.assert_not_called()

    # Assert job meta calls
    create_job_calls = [
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_COMPLETED}",
            job_type=JobType.BATCH,
            location=completed_task_location,
            input_dataset=NO_INPUTS,
            output_dataset=NO_OUTPUTS,
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            namespace_name=DAG_NAMESPACE
        ),
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_FAILED}",
            job_type=JobType.BATCH,
            location=failed_task_location,
            input_dataset=NO_INPUTS,
            output_dataset=NO_OUTPUTS,
            context=mock.ANY,
            description=DAG_DESCRIPTION,
            namespace_name=DAG_NAMESPACE
        )
    ]
    mock_marquez_client.create_job.assert_has_calls(create_job_calls)

    # Assert job run meta calls
    create_job_run_calls = [
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_COMPLETED}",
            run_id=mock.ANY,
            run_args=DAG_RUN_ARGS,
            nominal_start_time=mock.ANY,
            nominal_end_time=mock.ANY,
            namespace_name=DAG_NAMESPACE
        ),
        mock.call(
            job_name=f"{DAG_ID}.{TASK_ID_FAILED}",
            run_id=mock.ANY,
            run_args=DAG_RUN_ARGS,
            nominal_start_time=mock.ANY,
            nominal_end_time=mock.ANY,
            namespace_name=DAG_NAMESPACE
        )
    ]
    mock_marquez_client.create_job_run.assert_has_calls(create_job_run_calls)

    # Assert start run meta calls
    start_job_run_calls = [
        mock.call(run_id=mock.ANY),
        mock.call(run_id=mock.ANY)
    ]
    mock_marquez_client.mark_job_run_as_started.assert_has_calls(
        start_job_run_calls
    )

    # (5) Start task that will be marked as completed
    task_will_complete.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

    DAG.handle_callback(dagrun, success=True, session=session)
    mock_marquez_client.mark_job_run_as_completed.assert_called_once_with(
        run_id=mock.ANY
    )

    # When a task run completes, the task outputs are also updated in order
    # to link a job version (=task version) to a dataset version.
    # Using a DummyOperator, no outputs exists, so assert that the create
    # dataset call is not invoked.
    mock_marquez_client.create_dataset.assert_not_called()

    # (6) Start task that will be marked as failed
    task_will_fail.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

    DAG.handle_callback(dagrun, success=False, session=session)
    mock_marquez_client.mark_job_run_as_failed.assert_called_once_with(
        run_id=mock.ANY
    )

    # Assert an attempt to version the outputs of a task is not made when
    # a task fails
    mock_marquez_client.create_dataset.assert_not_called()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('VERSION'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as f:
    long_description = f.read()

NAME = "marquez-airflow"

setup(
    name=NAME,
    version=get_version('marquez_airflow/version.py'),
    author="Marquez Team",
    author_email="",
    description="Marquez integration with Airflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MarquezProject/marquez-airflow",
    packages=find_packages(),
    install_requires=[
        "marquez-python==0.7.8",
        "sqlparse==0.3.1"
    ],
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json

from google.cloud import bigquery

from marquez_airflow.utils import get_job_name
from airflow.contrib.operators.bigquery_operator import BigQueryOperator

from marquez_airflow.extractors.sql.experimental.parser import SqlParser
from marquez_airflow.extractors import (
    BaseExtractor,
    StepMetadata,
    Source,
    Dataset
)

# BIGQUERY DAGs doesn't use this.
# Required to pass marquez server validation.
_BIGQUERY_CONN_URL = \
    'jdbc:bigquery://https://www.googleapis.com/bigquery/v2:443'

log = logging.getLogger(__name__)


class BigQueryExtractor(BaseExtractor):
    operator_class = BigQueryOperator

    def __init__(self, operator):
        super().__init__(operator)

    def extract(self) -> [StepMetadata]:
        sql_meta = SqlParser.parse(self.operator.sql)
        log.info("bigquery sql parse successful.")

        # default value: 'bigquery_default'
        conn_id = self.operator.bigquery_conn_id
        source = Source(
            type="BIGQUERY",
            name=conn_id,
            connection_url=_BIGQUERY_CONN_URL)
        inputs = [
            Dataset.from_table(source, table) for table in sql_meta.in_tables
        ]
        outputs = [
            Dataset.from_table(source, table) for table in sql_meta.out_tables
        ]

        return [StepMetadata(
            name=get_job_name(task=self.operator),
            inputs=inputs,
            outputs=outputs,
            context={
                'sql': self.operator.sql
            }
        )]

    def extract_on_complete(self, task_instance) -> [StepMetadata]:
        steps_meta = self.extract()

        client = bigquery.Client()

        bigquery_job_id = task_instance.job_id
        job = client.get_job(job_id=bigquery_job_id)

        job_details = json.dumps(job._properties)
        log.debug(job_details)

        steps_meta[0].context = {
            'sql': self.operator.sql,
            'bigquery.job_details': job_details
        }

        client.close()

        return steps_meta



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from unittest.mock import patch

import pytest
from tests.mocks.git_mock import execute_git_mock

from marquez_airflow.utils import get_location
log = logging.getLogger(__name__)


@patch('marquez_airflow.utils.execute_git',
       side_effect=execute_git_mock)
def test_dag_location(git_mock):
    assert ('https://github.com/MarquezProject/marquez-airflow/blob/'
            'abcd1234/tests/test_dags/test_dag.py' ==
            get_location("tests/test_dags/test_dag.py"))


@patch('marquez_airflow.utils.execute_git',
       side_effect=execute_git_mock)
def test_bad_file_path(git_mock):
    log.debug("test_bad_file_path()")
    with pytest.raises(FileNotFoundError):
        # invalid file
        get_location("dags/missing-dag.py")


if __name__ == "__main__":
    pytest.main([sys.argv[0]])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from marquez_airflow.utils import execute_git

log = logging.getLogger(__name__)


def execute_git_mock(*args, **kwargs):
    # just mock the git revision
    log.debug("execute_git_mock()")
    if len(args) == 2 and len(args[1]) > 0 and args[1][0] == 'rev-list':
        return 'abcd1234'
    return execute_git(*args, **kwargs)
